@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2018 Antoine Martin <antoine@xpra.org>
+# Copyright (C) 2010-2020 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
@@ -10,7 +10,7 @@ from xpra.util import csv, nonl, envbool
 from xpra.os_util import bytestostr
 from xpra.gtk_common.keymap import get_gtk_keymap
 from xpra.gtk_common.gtk_util import keymap_get_for_display, display_get_default, get_default_root_window
-from xpra.gtk_common.error import xsync
+from xpra.gtk_common.error import xsync, xlog
 from xpra.gtk_common.gobject_compat import import_gdk, is_gtk3
 from xpra.keyboard.mask import DEFAULT_MODIFIER_NUISANCE, DEFAULT_MODIFIER_NUISANCE_KEYNAMES, mask_to_names
 from xpra.server.keyboard_config_base import KeyboardConfigBase
@@ -20,6 +20,7 @@ from xpra.x11.xkbhelper import (
     get_modifiers_from_meanings, get_modifiers_from_keycodes,
     clear_modifiers, set_modifiers, map_missing_modifiers,
     clean_keyboard_state,
+    DEBUG_KEYSYMS,
     )
 from xpra.x11.bindings.keyboard_bindings import X11KeyboardBindings #@UnresolvedImport
 from xpra.log import Logger
@@ -29,6 +30,7 @@ log = Logger("keyboard")
 X11Keyboard = X11KeyboardBindings()
 
 MAP_MISSING_MODIFIERS = envbool("XPRA_MAP_MISSING_MODIFIERS", True)
+SIMULATE_MODIFIERS = envbool("XPRA_SIMULATE_MODIFIERS", True)
 
 ALL_X11_MODIFIERS = {
                     "shift"     : 0,
@@ -69,6 +71,7 @@ class KeyboardConfig(KeyboardConfigBase):
         self.modifier_client_keycodes = {}
         self.compute_modifier_map()
         self.modifiers_filter = []
+        self.keycode_mappings = {}
 
     def __repr__(self):
         return "KeyboardConfig(%s / %s / %s)" % (self.xkbmap_layout, self.xkbmap_variant, self.xkbmap_options)
@@ -149,7 +152,7 @@ class KeyboardConfig(KeyboardConfigBase):
         parse_option("keycodes", props.listget)
         #dicts:
         for x in ("mod_meanings", "x11_keycodes", "query_struct"):
-            parse_option(x, props.dictget)
+            parse_option(x, props.dictget, {})
         #lists of strings:
         for x in ("mod_managed", "mod_pointermissing"):
             parse_option(x, props.strlistget)
@@ -311,66 +314,64 @@ class KeyboardConfig(KeyboardConfigBase):
     def set_keymap(self, translate_only=False):
         if not self.enabled:
             return
-        log("set_keymap(%s) layout=%s, variant=%s, options=%s, print=%s, query=%s", translate_only, self.xkbmap_layout, self.xkbmap_variant, self.xkbmap_options, nonl(self.xkbmap_print), nonl(self.xkbmap_query))
+        log("set_keymap(%s) layout=%s, variant=%s, options=%s, print=%s, query=%s",
+            translate_only, self.xkbmap_layout, self.xkbmap_variant, self.xkbmap_options,
+            nonl(self.xkbmap_print), nonl(self.xkbmap_query))
         if translate_only:
             self.keycode_translation = set_keycode_translation(self.xkbmap_x11_keycodes, self.xkbmap_keycodes)
             self.add_gtk_keynames()
             self.compute_modifiers()
             self.compute_modifier_keynames()
             self.compute_client_modifier_keycodes()
+            self.update_keycode_mappings()
             return
 
-        try:
-            with xsync:
-                clean_keyboard_state()
-                do_set_keymap(self.xkbmap_layout, self.xkbmap_variant, self.xkbmap_options,
-                              self.xkbmap_print, self.xkbmap_query, self.xkbmap_query_struct)
-        except:
-            log.error("Error setting up new keymap", exc_info=True)
+        with xlog:
+            clean_keyboard_state()
+            do_set_keymap(self.xkbmap_layout, self.xkbmap_variant, self.xkbmap_options,
+                          self.xkbmap_print, self.xkbmap_query, self.xkbmap_query_struct)
         log("set_keymap: xkbmap_print=%s, xkbmap_query=%s", nonl(self.xkbmap_print), nonl(self.xkbmap_query))
-        try:
-            with xsync:
-                #first clear all existing modifiers:
+        with xlog:
+            #first clear all existing modifiers:
+            clean_keyboard_state()
+
+            if not self.xkbmap_raw:
+                has_keycodes = bool(self.xkbmap_x11_keycodes) or bool(self.xkbmap_keycodes)
+                assert has_keycodes, "client failed to provide any keycodes!"
+
+                clear_modifiers(ALL_X11_MODIFIERS.keys())       #just clear all of them (set or not)
                 clean_keyboard_state()
 
-                if not self.xkbmap_raw:
-                    has_keycodes = bool(self.xkbmap_x11_keycodes) or bool(self.xkbmap_keycodes)
-                    assert has_keycodes, "client failed to provide any keycodes!"
-
-                    clear_modifiers(ALL_X11_MODIFIERS.keys())       #just clear all of them (set or not)
-                    clean_keyboard_state()
-
-                    #now set all the keycodes:
-                    #first compute the modifier maps as this may have an influence
-                    #on the keycode mappings (at least for the from_keycodes case):
-                    self.compute_modifiers()
-                    #key translation:
-                    if bool(self.xkbmap_query):
-                        #native full mapping of all keycodes:
-                        self.keycode_translation = set_all_keycodes(self.xkbmap_x11_keycodes, self.xkbmap_keycodes, False, self.keynames_for_mod)
-                    else:
-                        #if the client does not provide a full native keymap with all the keycodes,
-                        #try to preserve the initial server keycodes and translate the client keycodes instead:
-                        #(used by non X11 clients like osx,win32 or HTML5)
-                        self.keycode_translation = set_keycode_translation(self.xkbmap_x11_keycodes, self.xkbmap_keycodes)
-                    self.add_gtk_keynames()
-
-                    #now set the new modifier mappings:
-                    clean_keyboard_state()
-                    log("going to set modifiers, xkbmap_mod_meanings=%s, len(xkbmap_keycodes)=%s, keynames_for_mod=%s", self.xkbmap_mod_meanings, len(self.xkbmap_keycodes or []), self.keynames_for_mod)
-                    if self.keynames_for_mod:
-                        set_modifiers(self.keynames_for_mod)
-                    log("keynames_for_mod=%s", self.keynames_for_mod)
-                    self.compute_modifier_keynames()
+                #now set all the keycodes:
+                #first compute the modifier maps as this may have an influence
+                #on the keycode mappings (at least for the from_keycodes case):
+                self.compute_modifiers()
+                #key translation:
+                if bool(self.xkbmap_query):
+                    #native full mapping of all keycodes:
+                    self.keycode_translation = set_all_keycodes(self.xkbmap_x11_keycodes, self.xkbmap_keycodes, False, self.keynames_for_mod)
                 else:
-                    self.keycode_translation = {}
-                    log("keyboard raw mode, keycode translation left empty")
-                    self.compute_modifiers()
-                self.compute_client_modifier_keycodes()
-                log("keyname_for_mod=%s", self.keynames_for_mod)
+                    #if the client does not provide a full native keymap with all the keycodes,
+                    #try to preserve the initial server keycodes and translate the client keycodes instead:
+                    #(used by non X11 clients like osx,win32 or HTML5)
+                    self.keycode_translation = set_keycode_translation(self.xkbmap_x11_keycodes, self.xkbmap_keycodes)
+                self.add_gtk_keynames()
+
+                #now set the new modifier mappings:
                 clean_keyboard_state()
-        except:
-            log.error("Error setting X11 keymap", exc_info=True)
+                log("going to set modifiers, xkbmap_mod_meanings=%s, len(xkbmap_keycodes)=%s, keynames_for_mod=%s", self.xkbmap_mod_meanings, len(self.xkbmap_keycodes or []), self.keynames_for_mod)
+                if self.keynames_for_mod:
+                    set_modifiers(self.keynames_for_mod)
+                log("keynames_for_mod=%s", self.keynames_for_mod)
+                self.compute_modifier_keynames()
+            else:
+                self.keycode_translation = {}
+                log("keyboard raw mode, keycode translation left empty")
+                self.compute_modifiers()
+            self.compute_client_modifier_keycodes()
+            log("keyname_for_mod=%s", self.keynames_for_mod)
+            clean_keyboard_state()
+            self.update_keycode_mappings()
 
     def add_gtk_keynames(self):
         #add the keynames we find via gtk
@@ -379,6 +380,19 @@ class KeyboardConfig(KeyboardConfigBase):
         for _, keyname, keycode, _, _ in get_gtk_keymap():
             if keyname not in self.keycode_translation:
                 self.keycode_translation[keyname] = keycode
+                if keyname in DEBUG_KEYSYMS:
+                    log.info("add_gtk_keynames: %s=%s", keyname, keycode)
+        #some keys may still be missing with GTK2:
+        for keyname, equiv in {
+            "dead_perispomeni"  : "dead_tilde",
+            }.items():
+            if keyname not in self.keycode_translation:
+                keycode = self.keycode_translation.get(equiv)
+                if keycode:
+                    self.keycode_translation[keyname] = keycode
+                    if keyname in DEBUG_KEYSYMS:
+                        log.info("add_gtk_keynames: %s=%s", keyname, keycode)
+
 
     def set_default_keymap(self):
         """ assign a default keymap based on the current X11 server keymap
@@ -392,9 +406,21 @@ class KeyboardConfig(KeyboardConfigBase):
             #keycodes:
             keycode_to_keynames = X11Keyboard.get_keycode_mappings()
             self.keycode_translation = {}
+            #prefer keycodes that don't use the lowest level+mode:
+            default_for_keyname = {}
             for keycode, keynames in keycode_to_keynames.items():
-                for keyname in keynames:
-                    self.keycode_translation[keyname] = keycode
+                for i, keyname in enumerate(keynames):
+                    self.keycode_translation[(keyname, i)] = keycode
+                    if keyname in DEBUG_KEYSYMS:
+                        log.info("set_default_keymap: %s=%s", (keyname, i), keycode)
+                    kd = default_for_keyname.get(keyname)
+                    if kd is None or kd[1]>i:
+                        default_for_keyname[keyname] = (keycode, i)
+            for keyname, kd in default_for_keyname.items():
+                keycode = kd[0]
+                self.keycode_translation[keyname] = keycode
+                if keyname in DEBUG_KEYSYMS:
+                    log.info("set_default_keymap: %s=%s", keyname, keycode)
             self.add_gtk_keynames()
             log("set_default_keymap: keycode_translation=%s", self.keycode_translation)
             #modifiers:
@@ -413,20 +439,30 @@ class KeyboardConfig(KeyboardConfigBase):
             log("set_default_keymap: keynames_for_mod=%s", self.keynames_for_mod)
             log("set_default_keymap: keycodes_for_modifier_keynames=%s", self.keycodes_for_modifier_keynames)
             log("set_default_keymap: modifier_map=%s", self.modifier_map)
+            self.update_keycode_mappings()
+
+    def update_keycode_mappings(self):
+        self.keycode_mappings = X11Keyboard.get_keycode_mappings()
 
 
-    def do_get_keycode(self, client_keycode, keyname, pressed, modifiers):
+    def do_get_keycode(self, client_keycode, keyname, pressed, modifiers, group):
         if not self.enabled:
             log("ignoring keycode since keyboard is turned off")
-            return -1
+            return -1, group
+        if keyname=="0xffffff":
+            return -1, group
+        if self.xkbmap_raw:
+            return client_keycode, group
         keycode = None
         if self.xkbmap_query:
             keycode = self.keycode_translation.get((client_keycode, keyname)) or client_keycode
             log("get_keycode(%s, %s, %s)=%s (native keymap)", client_keycode, keyname, modifiers, keycode)
         else:
+            if SIMULATE_MODIFIERS:
+                return self.do_get_keycode_new(client_keycode, keyname, pressed, modifiers, group)
             #non-native: try harder to find matching keysym
             #first, try to honour shift state:
-            shift = "shift" in modifiers
+            shift = ("shift" in modifiers) ^ ("lock" in modifiers)
             mode = 0
             for mod in modifiers:
                 names = self.keynames_for_mod.get(mod, [])
@@ -435,11 +471,113 @@ class KeyboardConfig(KeyboardConfigBase):
                         mode = 1
                         break
             level = int(shift) + int(mode)*2
-            keycode = self.keycode_translation.get((keyname, level))
+            levels = []
+            if group:
+                levels.append(level+4)
+            levels.append(level)
+            for i in range(4):
+                level = int(shift) + i*2
+                if level not in levels:
+                    levels.append(level)
+            for level in levels:
+                keycode = self.keycode_translation.get((keyname, level))
+                if keycode:
+                    log("get_keycode(%s, %s, %s, %s)=%i (level=%i, shift=%s, mode=%i)",
+                        client_keycode, keyname, modifiers, group, keycode, level, shift, mode)
+                    break
             if keycode is None:
-                keycode = self.keycode_translation.get(keyname, client_keycode)
-            log("get_keycode(%s, %s, %s)=%i (level=%i)", client_keycode, keyname, modifiers, keycode, level)
-        return keycode
+                keycode = self.keycode_translation.get(keyname, -1)
+                log("get_keycode(%s, %s)=%i (keyname translation)", client_keycode, keyname, keycode)
+        return keycode, group
+
+    def do_get_keycode_new(self, client_keycode, keyname, pressed, modifiers, group):
+        #non-native: try harder to find matching keysym
+        def kmlog(msg, *args):
+            if keyname in DEBUG_KEYSYMS:
+                l = log.info
+            else:
+                l = log
+            l(msg, *args)
+        def klog(msg, *args):
+            kmlog("do_get_keycode%s"+msg, (client_keycode, keyname, pressed, modifiers, group), *args)
+        #first, try to honour shift state:
+        shift = ("shift" in modifiers) ^ ("lock" in modifiers)
+        mode = 0
+        numlock = 0
+        numlock_modifier = None
+        for mod, keynames in self.keynames_for_mod.items():
+            if "Num_Lock" in keynames:
+                numlock_modifier = mod
+                break
+        for mod in modifiers:
+            names = self.keynames_for_mod.get(mod, [])
+            if "Num_Lock" in names:
+                numlock = 1
+            for name in names:
+                if name in ("ISO_Level3_Shift", "Mode_switch"):
+                    mode = 1
+                    break
+        level = int(shift) + int(mode)*2
+        levels = []
+        #try to preserve the mode (harder to toggle):
+        for m in (int(bool(mode)), int(not mode)):
+            #try to preserve shift state:
+            for s in (int(bool(shift)), int(not shift)):
+                #group is comparatively easier to toggle (one function call):
+                for g in (int(bool(group)), int(not group)):
+                    level = int(g)*4 + int(m)*2 + int(s)*1
+                    levels.append(level)
+        kmlog("will try levels: %s", levels)
+        rgroup = group
+        for level in levels:
+            keycode = self.keycode_translation.get((keyname, level))
+            if keycode:
+                klog("=%i (level=%i, shift=%s, mode=%i)", keycode, level, shift, mode)
+                if self.xkbmap_raw:
+                    break
+                keysyms = self.keycode_mappings.get(keycode)
+                level0 = levels[0]
+                if len(keysyms)>level0 and keysyms[level0]=="":
+                    #if the keysym we would match for this keycode is 'NoSymbol',
+                    #then we can probably ignore it ('NoSymbol' shows up as "")
+                    kmlog("not toggling any modifiers state for keysyms=%s", keysyms)
+                    break
+                def toggle_modifier(mod):
+                    keynames = self.keynames_for_mod.get(mod)
+                    if keyname in keynames:
+                        kmlog("not toggling '%s' since '%s' should deal with it", mod, keyname)
+                        #the keycode we're returning is for this modifier,
+                        #assume that this will end up doing what is needed
+                        return
+                    if mod in modifiers:
+                        kmlog("removing '%s' from modifiers", mod)
+                        modifiers.remove(mod)
+                    else:
+                        kmlog("adding '%s' to modifiers", mod)
+                        modifiers.append(mod)
+                #keypad overrules shift state (see #2702):
+                if keyname.startswith("KP_"):
+                    if numlock_modifier and not numlock:
+                        toggle_modifier(numlock_modifier)
+                elif (level & 1) ^ shift:
+                    #shift state does not match
+                    toggle_modifier("shift")
+                if int(bool(level & 2)) ^ mode:
+                    #try to set / unset mode:
+                    for mod, keynames in self.keynames_for_mod.items():
+                        if "ISO_Level3_Shift" in keynames or "Mode_switch" in keynames:
+                            #found mode switch modified
+                            toggle_modifier(mod)
+                            break
+                rgroup = level//4
+                if rgroup!=group:
+                    kmlog("switching group from %i to %i", group, rgroup)
+                break
+        #this should not find anything new?:
+        if keycode is None:
+            keycode = self.keycode_translation.get(keyname, -1)
+            klog("=%i (keyname translation)", keycode)
+        return keycode, rgroup
 
 
     def get_current_mask(self):
