@@ -53,6 +53,7 @@ gllog = Logger("screen", "opengl")
 
 
 ALWAYS_NOTIFY_MOTION = envbool("XPRA_ALWAYS_NOTIFY_MOTION", False)
+FAKE_X11_INIT_ERROR = envbool("XPRA_FAKE_X11_INIT_ERROR", False)
 
 
 class XTestPointerDevice(object):
@@ -120,6 +121,8 @@ class X11ServerCore(GTKServerBase):
 
 
     def x11_init(self):
+        if FAKE_X11_INIT_ERROR:
+            raise Exception("fake x11 init error")
         clean_keyboard_state()
         if self.fake_xinerama in FALSE_OPTIONS:
             self.libfakeXinerama_so = None
@@ -395,7 +398,8 @@ class X11ServerCore(GTKServerBase):
             with xlog:
                 info.setdefault("keyboard", {}).update({
                     "state"             : {
-                        "keys_pressed"   : tuple(self.keys_pressed.keys())
+                        "keys_pressed"  : tuple(self.keys_pressed.keys()),
+                        "keycodes-down" : X11Keyboard.get_keycodes_down(),
                         },
                     "fast-switching"    : True,
                     "layout-group"      : X11Keyboard.get_layout_group(),
